//
//  JSExitModel.cpp
//  JSONDemo
//
//  This encapsulates all of the information for representing an exit door. Since the door
//  will never move, we normally would not need a separate class (because we could just use
//  an instance of BoxObstacle).  However, this changes when we do dynamic level loading.
//
//  The problem is that we need to initialize an physics object when a level is loaded,
//  but NOT attach any assets yet.  That is because the physics object is itself an asset
//  and it is not safe to assume the order of asset loading.  So instead, we just store
//  the keys of the assets in this class, and they attach the assets later when we
//  initialize the scene graph.
//
//  This file is based on the CS 3152 PhysicsDemo Lab by Don Holden, 2007
//
//  Author: Walker White
//  Version: 1/20/16
//
#include "JSExitModel.h"
#include <cornell/CUAssetManager.h>
#include <cornell/CUSceneManager.h>
#include <cornell/CUPolygonNode.h>


using namespace cocos2d;

#pragma mark -
#pragma mark Static Constructors
/**
 * Creates a new exit at the origin.
 *
 * The exit is 1 unit by 1 unit in size. The exit is scaled so that
 * 1 pixel = 1 Box2d unit
 *
 * The scene graph is completely decoupled from the physics system.
 * The node does not have to be the same size as the physics body. We
 * only guarantee that the scene graph node is positioned correctly
 * according to the drawing scale.
 *
 * Unlike most constructors, this one RETAINS the object and does not
 * autorelease.  That is because (1) this object is allocated in a
 * second thread (the asset loader thread) and (2) autoreleased objects
 * are freed in the main thread.  Hence autorelease is not thread safe,
 *
 * @retain a reference to the object (transferable)
 *
 * @return  An RETAINED physics object
 */
ExitModel* ExitModel::create() {
    ExitModel* exit = new (std::nothrow) ExitModel();
    if (exit && exit->init()) {
        // AUTORELEASE IS NOT THREAD SAFE FOR ASSET LOADING
        exit->retain();
        return exit;
    }
    CC_SAFE_DELETE(exit);
    return nullptr;
}

/**
 * Creates a new exit with the given position
 *
 * The exit is 1 unit by 1 unit in size. The exit is scaled so that
 * 1 pixel = 1 Box2d unit
 *
 * The scene graph is completely decoupled from the physics system.
 * The node does not have to be the same size as the physics body. We
 * only guarantee that the scene graph node is positioned correctly
 * according to the drawing scale.
 *
 * Unlike most constructors, this one RETAINS the object and does not
 * autorelease.  That is because (1) this object is allocated in a
 * second thread (the asset loader thread) and (2) autoreleased objects
 * are freed in the main thread.  Hence autorelease is not thread safe,
 *
 * @param  pos  Initial position in world coordinates
 *
 * @retain a reference to the object (transferable)
 *
 * @return  An RETAINED physics object
 */
ExitModel* ExitModel::create(const Vec2& pos) {
    ExitModel* exit = new (std::nothrow) ExitModel();
    if (exit && exit->init(pos)) {
        // AUTORELEASE IS NOT THREAD SAFE FOR ASSET LOADING
        exit->retain();
        return exit;
    }
    CC_SAFE_DELETE(exit);
    return nullptr;
}

/**
 * Creates a new exit with the given position and size.
 *
 * The exit size is specified in world coordinates.
 *
 * The scene graph is completely decoupled from the physics system.
 * The node does not have to be the same size as the physics body. We
 * only guarantee that the scene graph node is positioned correctly
 * according to the drawing scale.
 *
 * Unlike most constructors, this one RETAINS the object and does not
 * autorelease.  That is because (1) this object is allocated in a
 * second thread (the asset loader thread) and (2) autoreleased objects
 * are freed in the main thread.  Hence autorelease is not thread safe,
 *
 * @param  pos      Initial position in world coordinates
 * @param  size   	The dimensions of the exit door.
 *
 * @retain a reference to the object (transferable)
 *
 * @return  An autoreleased physics object
 */
ExitModel* ExitModel::create(const Vec2& pos, const Size& size) {
    ExitModel* exit = new (std::nothrow) ExitModel();
    if (exit && exit->init(pos,size)) {
        // AUTORELEASE IS NOT THREAD SAFE FOR ASSET LOADING
        exit->retain();
        return exit;
    }
    CC_SAFE_DELETE(exit);
    return nullptr;
}


#pragma mark -
#pragma mark Initializers
/**
 * Initializes a new exit with the given position and size.
 *
 * The exit size is specified in world coordinates.
 *
 * The scene graph is completely decoupled from the physics system.
 * The node does not have to be the same size as the physics body. We
 * only guarantee that the scene graph node is positioned correctly
 * according to the drawing scale.
 *
 * @param  pos      Initial position in world coordinates
 * @param  size   	The dimensions of the exit door.
 *
 * @return  true if the obstacle is initialized properly, false otherwise.
 */
bool ExitModel::init(const Vec2& pos, const Size& size) {
    BoxObstacle::init(pos,size);
    string name("goal");
    setName(name);
    setSensor(true);
    _exitTexture = "";
    return true;
}


#pragma mark -
#pragma mark Animation
/**
 * Performs any necessary additions to the scene graph node.
 *
 * This method is necessary for custom physics objects that are composed
 * of multiple scene graph nodes.  In this case, it is because we
 * manage our own afterburner animations.
 */
void ExitModel::resetSceneNode() {
    PolygonNode* pnode = dynamic_cast<PolygonNode*>(_node);
    if (pnode != nullptr) {
        // We need to know the content scale for resolution independence
        // If the device is higher resolution than 1024x576, Cocos2d will scale it
        // THIS DOES NOT FIX ASPECT RATIO PROBLEMS
        // If you are using a device with a 3:2 aspect ratio, you will need to
        // completely redo the level layout.  We can help if this is an issue.
        float cscale = Director::getInstance()->getContentScaleFactor();

        SceneManager* assets =  AssetManager::getInstance()->getCurrent();
        Texture2D* image = assets->get<Texture2D>(_exitTexture);
        
        Rect bounds;
        bounds.size = getDimension();
        bounds.size.width  *= _drawScale.x/cscale;
        bounds.size.height *= _drawScale.y/cscale;
        
        pnode->setTexture(image);
        pnode->setPolygon(bounds);
        pnode->setScale(cscale);
    }
}
