//
//  JSExitModel.h
//  JSONDemo
//
//  This encapsulates all of the information for representing an exit door. Since the door
//  will never move, we normally would not need a separate class (because we could just use
//  an instance of BoxObstacle).  However, this changes when we do dynamic level loading.
//
//  The problem is that we need to initialize an physics object when a level is loaded,
//  but NOT attach any assets yet.  That is because the physics object is itself an asset
//  and it is not safe to assume the order of asset loading.  So instead, we just store
//  the keys of the assets in this class, and they attach the assets later when we
//  initialize the scene graph.
//
//  This file is based on the CS 3152 PhysicsDemo Lab by Don Holden, 2007
//
//  Author: Walker White
//  Version: 1/20/16
//
#ifndef __JS_EXIT_MODEL_H__
#define __JS_EXIT_MODEL_H__

#include <cornell/CUBoxObstacle.h>

using namespace cocos2d;


#pragma mark -
#pragma mark Exit Door Model

/**
 * An exit door in the rocket lander game.
 *
 * Note that this class does keeps track of its own assets via keys, so that
 * they can be attached later as part of the scene graph.
 */
class ExitModel : public BoxObstacle {
private:
    /** This macro disables the copy constructor (not allowed on scene graphs) */
    CC_DISALLOW_COPY_AND_ASSIGN(ExitModel);
    
protected:
    /** The texture filmstrip for the exit door */
    std::string _exitTexture;
    
    /** The color for the debug outline */
    Color3B _debugColor;
    /** The opacity for the debug outline */
    GLubyte _debugOpacity;

    /**
     * Performs any necessary additions to the scene graph node.
     *
     * This method is necessary for custom physics objects that are composed
     * of multiple scene graph nodes.  In this case, it is because we
     * manage our own afterburner animations.
     */
    virtual void resetSceneNode() override;
    
    
#pragma mark -
#pragma mark Static Constructors
public:
    /**
     * Creates a new exit at the origin.
     *
     * The exit is 1 unit by 1 unit in size. The exit is scaled so that
     * 1 pixel = 1 Box2d unit
     *
     * The scene graph is completely decoupled from the physics system.
     * The node does not have to be the same size as the physics body. We
     * only guarantee that the scene graph node is positioned correctly
     * according to the drawing scale.
     *
     * Unlike most constructors, this one RETAINS the object and does not
     * autorelease.  That is because (1) this object is allocated in a
     * second thread (the asset loader thread) and (2) autoreleased objects
     * are freed in the main thread.  Hence autorelease is not thread safe,
     *
     * @retain a reference to the object (transferable)
     *
     * @return  An RETAINED physics object
     */
    static ExitModel* create();
    
    /**
     * Creates a new exit with the given position
     *
     * The exit is 1 unit by 1 unit in size. The exit is scaled so that
     * 1 pixel = 1 Box2d unit
     *
     * The scene graph is completely decoupled from the physics system.
     * The node does not have to be the same size as the physics body. We
     * only guarantee that the scene graph node is positioned correctly
     * according to the drawing scale.
     *
     * Unlike most constructors, this one RETAINS the object and does not
     * autorelease.  That is because (1) this object is allocated in a
     * second thread (the asset loader thread) and (2) autoreleased objects
     * are freed in the main thread.  Hence autorelease is not thread safe,
     *
     * @param  pos  Initial position in world coordinates
     *
     * @retain a reference to the object (transferable)
     *
     * @return  An RETAINED physics object
     */
    static ExitModel* create(const Vec2& pos);
    
    /**
     * Creates a new exit with the given position and size.
     *
     * The exit size is specified in world coordinates.
     *
     * The scene graph is completely decoupled from the physics system.
     * The node does not have to be the same size as the physics body. We
     * only guarantee that the scene graph node is positioned correctly
     * according to the drawing scale.
     *
     * Unlike most constructors, this one RETAINS the object and does not
     * autorelease.  That is because (1) this object is allocated in a
     * second thread (the asset loader thread) and (2) autoreleased objects
     * are freed in the main thread.  Hence autorelease is not thread safe,
     *
     * @param  pos      Initial position in world coordinates
     * @param  size   	The dimensions of the exit door.
     *
     * @retain a reference to the object (transferable)
     *
     * @return  An autoreleased physics object
     */
    static ExitModel* create(const Vec2& pos, const Size& size);
    
    
#pragma mark -
#pragma mark Animation
    /**
     * Returns the texture (key) for this rocket
     *
     * The value returned is not a Texture2D value.  Instead, it is a key for
     * accessing the texture from the asset manager.
     *
     * @return the texture (key) for this rocket
     */
    const std::string& getTextureKey() const { return _exitTexture; }
    
    /**
     * Returns the texture (key) for this rocket
     *
     * The value returned is not a Texture2D value.  Instead, it is a key for
     * accessing the texture from the asset manager.
     *
     * @param  strip    the texture (key) for this rocket
     */
    void setTextureKey(std::string strip) { _exitTexture = strip; }
    
    /**
     * Returns the color of the debug outline.
     *
     * The debug outlines traces the boundary of the physics fixtures.
     *
     * @return the color of the debug outline.
     */
    const Color3B& getDebugColor() const { return _debugColor; }
    
    /**
     * Sets the color of the debug outline.
     *
     * The debug outlines traces the boundary of the physics fixtures.
     *
     * @parma  color    the color of the debug outline.
     */
    void setDebugColor(const Color3B& color) { _debugColor = color; }
    
    /**
     * Returns the opacity of the debug outline.
     *
     * The debug outlines traces the boundary of the physics fixtures.
     *
     * @return the opacity of the debug outline.
     */
    GLubyte getDebugOpacity() const { return _debugOpacity; }
    
    /**
     * Sets the opacity of the debug outline.
     *
     * The debug outlines traces the boundary of the physics fixtures.
     *
     * @parma  opacity  the opacity of the debug outline.
     */
    void setDebugOpacity(GLubyte opacity) { _debugOpacity = opacity; }

    
#pragma mark -
#pragma mark Initializers
CC_CONSTRUCTOR_ACCESS:
    /*
     * Creates a new exit at the origin.
     */
    ExitModel(void) : BoxObstacle() { }
    
    /**
     * Destroys this exit, releasing all resources.
     */
    virtual ~ExitModel(void) {}
    
    /**
     * Initializes a new exit at the origin.
     *
     * The exit is 1 unit by 1 unit in size. The exit is scaled so that
     * 1 pixel = 1 Box2d unit
     *
     * The scene graph is completely decoupled from the physics system.
     * The node does not have to be the same size as the physics body. We
     * only guarantee that the scene graph node is positioned correctly
     * according to the drawing scale.
     *
     * @return  true if the obstacle is initialized properly, false otherwise.
     */
    virtual bool init() override { return init(Vec2::ZERO,Size::ZERO); }
    
    /**
     * Initializes a new exit with the given position
     *
     * The exit is 1 unit by 1 unit in size. The exit is scaled so that
     * 1 pixel = 1 Box2d unit
     *
     * The scene graph is completely decoupled from the physics system.
     * The node does not have to be the same size as the physics body. We
     * only guarantee that the scene graph node is positioned correctly
     * according to the drawing scale.
     *
     * @param  pos  Initial position in world coordinates
     *
     * @return  true if the obstacle is initialized properly, false otherwise.
     */
    virtual bool init(const Vec2& pos) override { return init(pos,Size(1,1)); }
    
    /**
     * Initializes a new exit with the given position and size.
     *
     * The exit size is specified in world coordinates.
     *
     * The scene graph is completely decoupled from the physics system.
     * The node does not have to be the same size as the physics body. We
     * only guarantee that the scene graph node is positioned correctly
     * according to the drawing scale.
     *
     * @param  pos      Initial position in world coordinates
     * @param  size   	The dimensions of the exit door.
     *
     * @return  true if the obstacle is initialized properly, false otherwise.
     */
    virtual bool init(const Vec2& pos, const Size& size) override;
    
};

#endif /* defined(__JS_EXIT_MODEL_H__) */