//
//  RDGameController.cpp
//  JSONDemo
//
//  This is the most important class in this demo.  This class manages the gameplay
//  for this demo.  It also handles collision detection. There is not much to do for
//  collisions; our WorldController class takes care of all of that for us.  This
//  controller mainly transforms input into gameplay.
//
//  This file is based on the CS 3152 PhysicsDemo Lab by Don Holden, 2007
//
//  Author: Walker White
//  Version: 1/20/16
//
// This is the root, so there are a lot of includes
#include <cornell.h>
#include <Box2D/Dynamics/b2World.h>
#include <Box2D/Dynamics/Contacts/b2Contact.h>
#include <Box2D/Collision/b2Collision.h>
#include "JSGameController.h"
#include "JSInputController.h"
#include "JSRocketModel.h"
#include "JSCrateModel.h"
#include "JSWallModel.h"
#include "JSExitModel.h"
#include "JSLevelModel.h"

#include <string>
#include <iostream>
#include <sstream>

// This is not part of cornell.h and SHOULD come last
#include <cornell/CUGenericLoader.h>


using namespace cocos2d;
using namespace std;

#pragma mark -
#pragma mark Hard-Coded Constants
/** The key for the font reference */
#define PRIMARY_FONT        "felt"
/** The message to display on a win */
#define WIN_MESSAGE         "VICTORY!"
/** The message to display on a level reset */
#define RESET_MESSAGE       "Reseting"
/** The color of the message to display on a win */
#define MESSAGE_COLOR       Color3B::YELLOW
/** The key for collisions sounds */
#define COLLISION_SOUND     "bump"
/** Threshold for generating sound on collision */
#define SOUND_THRESHOLD     3

#pragma mark -
#pragma mark Asset Constants
/** The JSON file that is the asset directory */
#define ASSET_DIRECTORY     "jsons/assets.json"
/** The key the source file */
#define FILE_KEY           "file"
/** The key indentifying textures to load */
#define TEXTURE_KEY         "textures"
/** The key indicating if a texture is wrapped */
#define TEXTURE_WRAP        "wrap"
/** The key indentifying sounds to load */
#define SOUNDS_KEY          "sounds"
/** The key indicating the sound volume to use */
#define SOUND_VOLUME        "volume"
/** The key indentifying fonts to load */
#define FONTS_KEY           "fonts"
/** The key indentifying the fonts size to use */
#define FONT_SIZE           "size"

/** The source for our level file */
#define LEVEL_ONE_FILE      "jsons/level.json"
/** The key for our loaded level */
#define LEVEL_ONE_KEY       "level1"

#pragma mark -
#pragma mark Initialization
/**
 * Creates a new game world with the default values.
 *
 * This constructor does not allocate any objects or start the controller.
 * This allows us to use a controller without a heap pointer.
 */
GameController::GameController() :
_rootnode(nullptr),
_level(nullptr),
_active(false),
_complete(false),
_debug(false)
{    
}

/**
 * Initializes the controller contents, and starts the game
 *
 * The constructor does not allocate any objects or memory.  This allows
 * us to have a non-pointer reference to this controller, reducing our
 * memory allocation.  Instead, allocation happens in this method.
 *
 * The game world is scaled so that the screen coordinates do not agree
 * with the Box2d coordinates.  This initializer uses the default scale.
 */
bool GameController::init(RootLayer* root) {
    _level = _assets->get<LevelModel>(LEVEL_ONE_KEY);
    if (_level == nullptr || root == nullptr) {
        return false;
    }
    
    _input.init();
    _input.start();
    
    _rootnode = root;
    _rootnode->retain();
    _level->setRootNode(root); // Obtains ownership of root.
    
    // Create the world and attach the listeners.
    WorldController* world = _level->getWorld();
    world->activateCollisionCallbacks(true);
    world->onBeginContact = [this](b2Contact* contact) {
        beginContact(contact);
    };
    world->beforeSolve = [this](b2Contact* contact, const b2Manifold* oldManifold) {
        beforeSolve(contact,oldManifold);
    };
    
    
    // Create the node for winning
    _winnode = Label::create();
    
    _winnode->setTTFConfig(_assets->get<TTFont>(PRIMARY_FONT)->getTTF());
    _winnode->setString(WIN_MESSAGE);
    _winnode->setColor(MESSAGE_COLOR);
    _winnode->setVisible(false);
    _winnode->setPosition(root->getContentSize().width/2.0f,root->getContentSize().height/2.0f);
    root->addChild(_winnode,3);
    
    // Create the node for reseting
    _loadnode = Label::create();
    
    _loadnode->setTTFConfig(_assets->get<TTFont>(PRIMARY_FONT)->getTTF());
    _loadnode->setString(RESET_MESSAGE);
    _loadnode->setColor(MESSAGE_COLOR);
    _loadnode->setVisible(false);
    _loadnode->setPosition(root->getContentSize().width/2.0f,root->getContentSize().height/2.0f);
    root->addChild(_loadnode,4);
    
    _active = true;
    _complete = false;
    setDebug(false);
    return true;
}

/**
 * Disposes of all (non-static) resources allocated to this mode.
 *
 * This method is different from dispose() in that it ALSO deletes the
 * input controller.
 */
GameController::~GameController() {
    dispose();
}

/**
 * Disposes of all (non-static) resources allocated to this mode.
 */
void GameController::dispose() {
    if (_level != nullptr) {
        _level->release();
        _level = nullptr;
    }
    if (_rootnode != nullptr) {
        _rootnode->release();
        _rootnode = nullptr;
    }
}


#pragma mark -
#pragma mark Physics Handling

/**
 * Executes the core gameplay loop of this world.
 *
 * This method contains the specific update code for this mini-game. It does
 * not handle collisions, as those are managed by the parent class WorldController.
 * This method is called after input is read, but before collisions are resolved.
 * The very last thing that it should do is apply forces to the appropriate objects.
 *
 * @param  delta    Number of seconds since last animation frame
 */
void GameController::update(float dt) {
    if (_level == nullptr) {
        return;
    }
    
    // Check to see if new level loaded yet
    if (_loadnode->isVisible()) {
        if (_assets->isComplete()) {
            // Release the old level permanently
            _level->release();
            
            // Access and initialize level
            _level = _assets->get<LevelModel>(LEVEL_ONE_KEY);
            _level->setRootNode(_rootnode);
            _level->showDebug(_debug);
            
            _loadnode->setVisible(false);
        } else {
            // Level is not loaded yet; refuse input
            return;
        }
    }
    
    _input.update(dt);
    
    // Process the toggled key commands
    if (_input.didReset()) {
        // Unload the level but keep in memory temporarily
        _level->retain();
        _assets->unload<LevelModel>(LEVEL_ONE_KEY);
        
        // Load a new level and quit update
        _loadnode->setVisible(true);
        _assets->loadAsync<LevelModel>(LEVEL_ONE_KEY,LEVEL_ONE_FILE);
        setComplete(false);
        return;
    }
    if (_input.didDebug()) {
        setDebug(!isDebug());
    }
    if (_input.didExit())  {
        _rootnode->shutdown();
        return;
    }

    // Apply the force to the rocket
    RocketModel* rocket = _level->getRocket();
    if (rocket != nullptr) {
        rocket->setFX(_input.getHorizontal() * rocket->getThrust());
        rocket->setFY(_input.getVertical() * rocket->getThrust());
        rocket->applyForce();

        // Animate the three burners
        updateBurner(rocket, RocketModel::Burner::MAIN,  rocket->getFY() > 1);
        updateBurner(rocket, RocketModel::Burner::LEFT,  rocket->getFX() > 1);
        updateBurner(rocket, RocketModel::Burner::RIGHT, rocket->getFX() < -1);
    }
    
    // Turn the physics engine crank.
    _level->getWorld()->update(dt);
}

/**
 * Updates that animation for a single burner
 *
 * This method is here instead of the the rocket model because of our philosophy
 * that models should always be lightweight.  Animation includes sounds and other
 * assets that we do not want to process in the model
 *
 * @param  burner   The rocket burner to animate
 * @param  on       Whether to turn the animation on or off
 */
void GameController::updateBurner(RocketModel* rocket, RocketModel::Burner burner, bool on) {
    string sound = rocket->getBurnerSound(burner);
    if (on) {
        rocket->animateBurner(burner, true);
        if (!SoundEngine::getInstance()->isActiveEffect(sound) && sound.size() > 0) {
            Sound* source = _assets->get<Sound>(sound);
            SoundEngine::getInstance()->playEffect(sound,source,true,_volume[sound]);
        }
    } else {
        rocket->animateBurner(burner, false);
        if (SoundEngine::getInstance()->isActiveEffect(sound)) {
            SoundEngine::getInstance()->stopEffect(sound);
        }
    }
    
}

/**
 * Processes the start of a collision
 *
 * This method is called when we first get a collision between two objects.  We use
 * this method to test if it is the "right" kind of collision.  In particular, we
 * use it to test if we make it to the win door.
 *
 * @param  contact  The two bodies that collided
 */
void GameController::beginContact(b2Contact* contact) {
    b2Body* body1 = contact->GetFixtureA()->GetBody();
    b2Body* body2 = contact->GetFixtureB()->GetBody();
    
    RocketModel* rocket = _level->getRocket();
    ExitModel* goalDoor = _level->getExit();
    
    // If we hit the "win" door, we are done
    if((body1->GetUserData() == rocket && body2->GetUserData() == goalDoor) ||
       (body1->GetUserData() == goalDoor && body2->GetUserData() == rocket)) {
        setComplete(true);
    }
}

/**
 * Handles any modifications necessary before collision resolution
 *
 * This method is called just before Box2D resolves a collision.  We use this method
 * to implement sound on contact, using the algorithms outlined in Ian Parberry's
 * "Introduction to Game Physics with Box2D".
 *
 * @param  contact  The two bodies that collided
 * @param  contact  The collision manifold before contact
 */
void GameController::beforeSolve(b2Contact* contact, const b2Manifold* oldManifold) {
    float speed = 0;

    // Use Ian Parberry's method to compute a speed threshold
    b2Body* body1 = contact->GetFixtureA()->GetBody();
    b2Body* body2 = contact->GetFixtureB()->GetBody();
    b2WorldManifold worldManifold;
    contact->GetWorldManifold(&worldManifold);
    b2PointState state1[2], state2[2];
    b2GetPointStates(state1, state2, oldManifold, contact->GetManifold());
    for(int ii =0; ii < 2; ii++) {
        if (state2[ii] == b2_addState) {
            b2Vec2 wp = worldManifold.points[0];
            b2Vec2 v1 = body1->GetLinearVelocityFromWorldPoint(wp);
            b2Vec2 v2 = body2->GetLinearVelocityFromWorldPoint(wp);
            b2Vec2 dv = v1-v2;
            speed = b2Dot(dv,worldManifold.normal);
        }
    }
    
    // Play a sound if above threshold
    if (speed > SOUND_THRESHOLD) {
        string key = ((Obstacle*)body1->GetUserData())->getName()+((Obstacle*)body2->GetUserData())->getName();
        Sound* source = _assets->get<Sound>(COLLISION_SOUND);
        SoundEngine::getInstance()->playEffect(key, source, false, _volume[COLLISION_SOUND]);
    }
}

/**
 * Preloads the assets needed for the game.
 *
 * This a new version of the method that makes use of our JSON reader.  It reads a "directory"
 * from a JSON file, and uses that to load the assets.  Then it passes another JSON to a 
 * LevelModel, so that it can load from that file.
 */
void GameController::preload() {
    _assets = AssetManager::getInstance()->getCurrent();
    
    JSONReader reader;
    reader.initWithFile(ASSET_DIRECTORY);
    if (!reader.startJSON()) {
        CCASSERT(false, "Failed to load asset directory");
        return;
    }
    
    // Process textures
    if (reader.startObject(TEXTURE_KEY) > 0) {
        TextureLoader* tloader = (TextureLoader*)_assets->access<Texture2D>();
        
        // Wrap settings for identified textures
        Texture2D::TexParams params;
        params.wrapS = GL_REPEAT;
        params.wrapT = GL_REPEAT;
        params.magFilter = GL_LINEAR;
        params.minFilter = GL_NEAREST;
        
        // Convert the object to an array so we can see keys and values
        int tsize = reader.startArray();
        for(int ii = 0; ii < tsize; ii++) {
            string key   = reader.getKey();
            
            // Unwrap the object
            if (reader.startObject()) {
                string value = reader.getString(FILE_KEY);
                bool wrap = reader.getBool(TEXTURE_WRAP);
                // Set wrap settings if appropriate
                if (wrap) {
                    tloader->loadAsync(key,value,params);
                } else {
                    tloader->loadAsync(key,value);
                }
            }
            
            reader.endObject();
            reader.advance();
        }
        reader.endArray();
    }
    reader.endObject();
    
    // Process sounds
    if (reader.startObject(SOUNDS_KEY) > 0) {
        // Convert the object to an array so we can see keys and values
        int ssize = reader.startArray();
        for(int ii = 0; ii < ssize; ii++) {
            string key   = reader.getKey();
            
            // Unwrap the object
            if (reader.startObject()) {
                string value = reader.getString(FILE_KEY);
                float vol = reader.getNumber(SOUND_VOLUME);
                _volume.emplace(key,vol); // We have to store volume in a different place
                _assets->loadAsync<Sound>(key, value);
            }
            
            reader.endObject();
            reader.advance();
        }
        reader.endArray();
    }
    reader.endObject();
    
    // Process fonts
    if (reader.startObject(FONTS_KEY) > 0) {
        FontLoader* floader = (FontLoader*)_assets->access<TTFont>();
        
        // Convert the object to an array so we can see keys and values
        int ssize = reader.startArray();
        for(int ii = 0; ii < ssize; ii++) {
            string key   = reader.getKey();
            
            // Unwrap the object
            if (reader.startObject()) {
                string value = reader.getString(FILE_KEY);
                float size = reader.getNumber(FONT_SIZE);
                floader->loadAsync(key, value, size);
            }
            
            reader.endObject();
            reader.advance();
        }
        reader.endArray();
    }
    reader.endObject();
    reader.endJSON();
    
    // Finally, load the level
    _assets->loadAsync<LevelModel>(LEVEL_ONE_KEY,LEVEL_ONE_FILE);
}

/**
 * Sets whether debug mode is active.
 *
 * If true, all objects will display their physics bodies.
 *
 * @param value whether debug mode is active.
 */
void GameController::setDebug(bool value) {
    _debug = value;
    if (_level != nullptr) {
        _level->showDebug(value);
    }
}
